"""Base Spout classes."""
from ipc import read_handshake, read_command, send_message, json, _stdout
from base import Component


class Spout(Component):
    """Base class for all streamparse spouts.  For more information on spouts,
    consult Storm's `Concepts documentation <http://storm.incubator.apache.org/documentation/Concepts.html>`_.
    """

    def initialize(self, storm_conf, context):
        """Called immediately after the initial handshake with Storm and before
        the main run loop. A good place to initialize connections to data
        sources.

        :param storm_conf: the Storm configuration for this Bolt.  This is the
                           configuration provided to the topology, merged in
                           with cluster configuration on the worker node.
        :type storm_conf: dict
        :param context: information about the component's place within the
                        topology such as: task IDs, inputs, outputs etc.
        :type context: dict
        """
        pass

    def ack(self, tup_id):
        """Called when a bolt acknowledges a tuple in the topology.

        :param tup_id: the ID of the tuple that has been fully acknowledged in
                       the topology.
        :type tup_id: str
        """
        pass

    def fail(self, tup_id):
        """Called when a tuple fails in the topology, a Spout can choose to
        emit the tuple again or ignore the fail.  Default is to ignore.

        :param tup_id: the ID of the tuple that has failed in the topology
                       either due to a bolt calling ``fail()`` or a tuple
                       timing out.
        :type tup_id: str
        """
        pass

    def next_tuple(self):
        raise NotImplementedError()

    def emit(self, tup, tup_id=None, stream=None, direct_task=None):
        """Emit a spout tuple message.

        :param tup: the tuple to send to Storm.  Should contain only
                    JSON-serializable data.
        :type tup: list
        :param tup_id: the ID for the tuple. Leave this blank for an
                       unreliable emit.
        :type tup_id: str
        :param stream: ID of the stream this tuple should be emitted to.
                       Leave empty to emit to the default stream.
        :type stream: str
        :param direct_task: the task to send the tuple to if performing a
                            direct emit.
        :type direct_task: int
        """
        if not isinstance(tup, list):
            raise TypeError('All tuples must be lists, received {!r} instead'
                            .format(type(tup)))

        msg = {'command': 'emit', 'tuple': tup}
        if tup_id is not None:
            msg['id'] = tup_id
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        send_message(msg)

    def emit_many(self, tuples, tup_id=None, stream=None, direct_task=None):
        """A more efficient way to send many tuples, dumps out all tuples to
        STDOUT instead of writing one at a time.

        :param tuples: a two-dimensional ``list`` representing the tuples to
                       send to Storm.  Tuples should contain only
                       JSON-serializable data.
        :type tuples: list
        :param tup_id: the id for the tuple.  Leave this blank for unreliable
                       emits.
        :type tup_id: str
        :param stream: ID of the stream these tuples should be emitted to.
                       Leave empty to emit to the default stream.
        :type stream: str
        :param direct_task: the task to send the tuple to if
                            performing a direct emit.
        :type direct_task: int
        """
        msg = {
            'command': 'emit',
        }
        if tup_id is not None:
            msg['id'] = tup_id
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        lines = []
        for tup in tuples:
            msg['tuple'] = tup
            lines.append(json.dumps(msg))
        _stdout.write("{}\nend\n".format("\nend\n".join(lines)))
        _stdout.flush()

    def run(self):
        """Main run loop for all spouts. Performs initial handshake with Storm
        and reads tuples handing them off to subclasses.  Any exceptions are
        caught and logged back to Storm prior to the Python process exits.
        Subclasses should not override this method.
        """
        storm_conf, context = read_handshake()
        try:
            self.initialize(storm_conf, context)
            while True:
                cmd = read_command()
                if cmd['command'] == 'next':
                    self.next_tuple()
                if cmd['command'] == 'ack':
                    self.ack(cmd['id'])
                if cmd['command'] == 'fail':
                    self.fail(cmd['id'])
                send_message({'command': 'sync'})
        except Exception as e:
            self.raise_exception(e)



import sys
import os

here = os.path.split(os.path.abspath(__file__))[0]
root = os.path.abspath(os.path.join(here, '../../'))
sys.path[0:0] = [root]

from streamparse.bolt import BatchingBolt


class DummyBatchingBolt(BatchingBolt):

    def group_key(self, tup):
        return tup.values[0]

    def process_batch(self, key, tups):
        for tup in tups:
            if tup.values[0] == "fail":
                raise Exception("Something bad happened!")
            self.emit([key, tup.id])


if __name__ == '__main__':
    DummyBatchingBolt().run()



"""Base Bolt classes."""
from collections import defaultdict
import os
import signal
import sys
import threading
import time

from base import Component
from ipc import read_handshake, read_tuple, send_message, json, _stdout, Tuple


_ANCHOR_TUPLE = None


class Bolt(Component):
    """The base class for all streamparse bolts. For more information on bolts,
    consult Storm's `Concepts documentation <http://storm.incubator.apache.org/documentation/Concepts.html>`_.
    """

    def initialize(self, storm_conf, context):
        """Called immediately after the initial handshake with Storm and before
        the main run loop. A good place to initialize connections to data
        sources.

        :param storm_conf: the Storm configuration for this Bolt.  This is the
                           configuration provided to the topology, merged in
                           with cluster configuration on the worker node.
        :type storm_conf: dict
        :param context: information about the component's place within the
                        topology such as: task IDs, inputs, outputs etc.
        :type context: dict
        """
        pass

    def process(self, tup):
        """Process a single tuple :class:`Tuple` of input, should be
        overridden by subclasses.  :class:`Tuple` objects contain metadata
        about which component, stream and task it came from.  The actual values
        of the tuple can be accessed by calling ``tup.values``.

        :param tup: the tuple to be processed.
        :type tup: Tuple
        """
        raise NotImplementedError()

    def emit(self, tup, stream=None, anchors=[], direct_task=None):
        """Emit a new tuple to a stream.

        :param tup: the Tuple payload to send to Storm, should contain only
                    JSON-serializable data.
        :type tup: list
        :param stream: the ID of the stream to emit this tuple to. Specify
                       ``None`` to emit to default stream.
        :type stream: str
        :param anchors: IDs of the tuples the emitted tuple should be anchored
                        to.
        :type anchors: list
        :param direct_task: the task to send the tuple to.
        :type direct_task: int
        """
        if not isinstance(tup, list):
            raise TypeError('All tuples must be lists, received {!r} instead'
                            .format(type(tup)))
        if _ANCHOR_TUPLE is not None:
            anchors = [_ANCHOR_TUPLE]
        msg = {'command': 'emit', 'tuple': tup}
        if stream is not None:
            msg['stream'] = stream
        msg['anchors'] = [x.id for x in anchors]
        if direct_task is not None:
            msg['task'] = direct_task

        send_message(msg)

    def emit_many(self, tuples, stream=None, anchors=[], direct_task=None):
        """A more efficient way to send many tuples, dumps out all tuples to
        STDOUT instead of writing one at a time.

        :param tuples: a ``list``s containing ``list``s of tuple payload data
                       to send to Storm. All tuples should contain only
                       JSON-serializable data.
        :type tuples: list
        :param stream: the ID of the steram to emit these tuples to. Specify
                       ``None`` to emit to default stream.
        :type stream: str
        :param anchors: IDs the tuples which the emitted tuples should be
                        anchored to.
        :type anchors: list
        :param direct_task: indicates the task to send the tuple to.
        :type direct_task: int
        """
        if _ANCHOR_TUPLE is not None:
            anchors = [_ANCHOR_TUPLE]
        msg = {
            'command': 'emit',
            'anchors': [a.id for a in anchors],
        }
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        lines = []
        for tup in tuples:
            msg['tuple'] = tup
            lines.append(json.dumps(msg))
        _stdout.write("{}\nend\n".format("\nend\n".join(lines)))
        _stdout.flush()

    def ack(self, tup):
        """Indicate that processing of a tuple has succeeded.

        :param tup: the tuple to acknowledge.
        :type tup: str or Tuple
        """
        tup_id = tup.id if isinstance(tup, Tuple) else tup
        send_message({'command': 'ack', 'id': tup_id})

    def fail(self, tup):
        """Indicate that processing of a tuple has failed.

        :param tup: the tuple to fail.
        :type tup: str or Tuple
        """
        tup_id = tup.id if isinstance(tup, Tuple) else tup
        send_message({'command': 'fail', 'id': tup_id})

    def run(self):
        """Main run loop for all bolts. Performs initial handshake with Storm
        and reads tuples handing them off to subclasses.  Any exceptions are
        caught and logged back to Storm prior to the Python process exits.
        Subclasses should not override this method.
        """
        storm_conf, context = read_handshake()
        tup = None
        try:
            self.initialize(storm_conf, context)
            while True:
                tup = read_tuple()
                self.process(tup)
        except Exception as e:
            self.raise_exception(e, tup)


class BasicBolt(Bolt):
    """A bolt implementation that automatically acknowledges all tuples after
    :func:`process` completes."""

    def run(self):
        global _ANCHOR_TUPLE
        storm_conf, context = read_handshake()
        tup = None
        try:
            self.initialize(storm_conf, context)
            while True:
                tup = read_tuple()
                _ANCHOR_TUPLE = tup
                self.process(tup)
                self.ack(tup)
        except Exception as e:
            self.raise_exception(e, tup)


class BatchingBolt(Bolt):
    """A bolt which batches tuples for processing.

    Batching tuples is unexpectedly complex to do correctly. The main problem
    is that all bolts are single-threaded. The difficult comes when the topology
    is shutting down because Storm stops feeding the bolt tuples. If the bolt
    is blocked waiting on stdin, then it can't process any waiting tuples,
    or even ack ones that were asynchronously written to a data store.

    This bolt helps with that grouping tuples based on a time interval and then
    processing them on a worker thread. The bolt also handles ack'ing tuples
    after processing has finished, much like the BasicBolt.

    To use this class, you must implement ``process_batch``. ``group_key`` can
    be optionally implemented so that tuples are grouped before
    ``process_batch`` is even called.
    """
    SECS_BETWEEN_BATCHES = 2

    def __init__(self):
        self.exc_info = None
        signal.signal(signal.SIGINT, self._handle_worker_exception)

        self._batch = defaultdict(list)
        self._should_stop = threading.Event()
        self._batcher = threading.Thread(target=self._batcher)
        self._batch_lock = threading.Lock()
        self._batcher.daemon = True
        self._batcher.start()

    def process(self, tup):
        """Add a tuple a specific batch by group key. Do not override this
        method in subclasses.
        """
        with self._batch_lock:
            group_key = self.group_key(tup)
            self._batch[group_key].append(tup)

    def group_key(self, tup):
        """Return the group key used to group tuples within a batch.

        By default, returns None, which put all tuples in a single
        batch. Override this function to enable batching.

        :param tup: the tuple used to extract a group key
        :type tup: Tuple
        :returns: Any ``hashable`` value (will be used in a ``dict``).
        """
        return None

    def process_batch(self, key, tups):
        """Process a batch of tuples. Should be overridden by subclasses.

        :param key: the group key for the list of batches.
        :type key: hashable
        :param tups: a `list` of :class:`ipc.Tuple`s for the group.
        :type tups: list
        """
        raise NotImplementedError()

    def _batcher(self):
        """Entry point for the batcher thread."""
        try:
            while True:
                time.sleep(self.SECS_BETWEEN_BATCHES)
                with self._batch_lock:
                    if not self._batch:
                        # No tuples to save
                        continue
                    for key, tups in self._batch.iteritems():
                        self.process_batch(key, tups)
                        [self.ack(tup) for tup in tups]
                    self._batch = defaultdict(list)
        except Exception:
            self.exc_info = sys.exc_info()
            os.kill(os.getpid(), signal.SIGINT)  # interrupt stdin waiting

    def _handle_worker_exception(self, signum, frame):
        """Handle an exception raised in the worker thread.

        Exceptions in the _batcher thread will send a SIGINT to the main
        thread which we catch here, and then raise in the main thread.
        """
        raise self.exc_info[1], None, self.exc_info[2]



import sys
import os

here = os.path.split(os.path.abspath(__file__))[0]
root = os.path.abspath(os.path.join(here, '../../'))
sys.path[0:0] = [root]

from streamparse.spout import Spout


class DummySpout(Spout):

    def next_tuple(self):
        self.emit(["test"])


if __name__ == '__main__':
    DummySpout().run()



from __future__ import print_function, absolute_import
import subprocess
import unittest
import os
import sys
import time

from .util import ShellProcess


_ROOT = os.path.dirname(os.path.realpath(__file__))
def here(*x):
    return os.path.join(_ROOT, *x)


class ShellSpoutTester(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        args = ["python", here("dummy_spout.py")]
        cls.proc = subprocess.Popen(args, stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        print("Waiting for subprocess to start...")
        time.sleep(1)  # time for the subprocess to start
        if cls.proc.poll() is not None:
            raise Exception("Could not create subprocess.\n{}"
                            .format("".join(cls.proc.stderr.readlines())))
        cls.shell_proc = ShellProcess(cls.proc.stdout, cls.proc.stdin)

    def test_1_initial_handshake(self):
        msg = {
            "conf": {},
            "context": {},
            "pidDir": here()
        }
        ShellSpoutTester.shell_proc.write_message(msg)
        res = ShellSpoutTester.shell_proc.read_message()

        self.assertIsInstance(res, dict)
        self.assertEqual(res.get("pid"), ShellSpoutTester.proc.pid)
        pid = str(res["pid"])
        self.assertTrue(os.path.exists(here(pid)))
        self.assertTrue(os.path.isfile(here(pid)))

    def test_2_next_tuple(self):
        msg = {"command": "next"}
        res = ShellSpoutTester.shell_proc.query_subprocess(msg)

        self.assertEqual(len(res), 2)

        cmd = res[0]
        self.assertDictEqual({"command": "emit", "tuple": ["test"]}, cmd)

        cmd = res[1]
        self.assertDictEqual({"command": "sync"}, cmd)

    def test_3_ack(self):
        msg = {"command": "ack", "id": "123456"}
        res = ShellSpoutTester.shell_proc.query_subprocess(msg)

        self.assertEqual(len(res), 1)

        cmd = res[0]
        self.assertDictEqual({"command": "sync"}, cmd)

    def test_4_fail(self):
        msg = {"command": "fail", "id": "123456"}
        res = ShellSpoutTester.shell_proc.query_subprocess(msg)

        self.assertEqual(len(res), 1)

        cmd = res[0]
        self.assertDictEqual({"command": "sync"}, cmd)

    @classmethod
    def tearDownClass(cls):
        os.remove(here(str(cls.proc.pid)))
        cls.proc.kill()

if __name__ == '__main__':
    unittest.main()



import sys
import os

here = os.path.split(os.path.abspath(__file__))[0]
root = os.path.abspath(os.path.join(here, '../../'))
sys.path[0:0] = [root]

from streamparse.bolt import Bolt


class DummyBolt(Bolt):

    def process(self, tup):
        if tup.id != "emit_many":
            self.emit(tup.values)

        if tup.id == "ack_me":
            self.ack(tup)
        elif tup.id == "fail_me":
            self.fail(tup)
        elif tup.id == "emit_many":
            self.emit_many([tup.values]*5)


if __name__ == '__main__':
    DummyBolt().run()



from __future__ import print_function, absolute_import
import subprocess
import unittest
import os
import sys
import time

from .util import ShellProcess

_ROOT = os.path.dirname(os.path.realpath(__file__))
def here(*x):
    return os.path.join(_ROOT, *x)


class ShellBatchingBoltTester(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        args = ["python", here("dummy_batching_bolt.py")]
        cls.proc = subprocess.Popen(args, stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        print("Waiting for subprocess to start...")
        time.sleep(1)  # time for the subprocess to start
        if cls.proc.poll() is not None:
            raise Exception("Could not create subprocess.\n{}"
                            .format("".join(cls.proc.stderr.readlines())))
        cls.shell_proc = ShellProcess(cls.proc.stdout, cls.proc.stdin)

    def test_1_initial_handshake(self):
        msg = {
            "conf": {},
            "context": {},
            "pidDir": here()
        }
        ShellBatchingBoltTester.shell_proc.write_message(msg)
        res = ShellBatchingBoltTester.shell_proc.read_message()

        self.assertIsInstance(res, dict)
        self.assertEqual(res.get("pid"), ShellBatchingBoltTester.proc.pid)
        pid = str(res["pid"])
        self.assertTrue(os.path.exists(here(pid)))
        self.assertTrue(os.path.isfile(here(pid)))

    def test_2_ack_tuple(self):
        msg = {
            "id": "2285924946354050200",
            "comp": "word-spout",
            "stream": "default",
            "task": 0,
            "tuple": ["snow white and the seven dwarfs", "field2", 3, 4.252],
        }
        ShellBatchingBoltTester.shell_proc.write_message(msg)

        res = ShellBatchingBoltTester.shell_proc.read_message()
        self.assertEqual(res.get("command"), "emit")
        self.assertEqual(res.get("tuple"),
                         ["snow white and the seven dwarfs", msg["id"]])

        res = ShellBatchingBoltTester.shell_proc.read_message()
        self.assertEqual(res.get("command"), "ack")
        self.assertEqual(res.get("id"), msg["id"])

    def test_3_multi_ack(self):
        msg = {
            "id": None,
            "comp": "word-spout",
            "stream": "default",
            "task": 0,
            "tuple": ["snow white and the seven dwarfs", "field2", 3, 4.252],
        }
        for i in xrange(5):
            msg["id"] = str(i)
            ShellBatchingBoltTester.shell_proc.write_message(msg)
        time.sleep(2.5)
        # Ensure that we have a series of emits followed by a series of acks
        for i in xrange(5):
            res = ShellBatchingBoltTester.shell_proc.read_message()
            self.assertEqual(res.get("command"), "emit")
            self.assertEqual(res.get("tuple"),
                             ["snow white and the seven dwarfs",str(i)])
        for i in xrange(5):
            res = ShellBatchingBoltTester.shell_proc.read_message()
            self.assertEqual(res.get("command"), "ack")
            self.assertEqual(res.get("id"), str(i))

    def test_4_batches(self):
        # Create two batches the "mike" batch and the "lida" batch
        messages = (
            {
                "id": "0",
                "comp": "word-spout",
                "stream": "default",
                "task": 0,
                "tuple": ["mike", "smith"],
            },
            {
                "id": "1",
                "comp": "word-spout",
                "stream": "default",
                "task": 0,
                "tuple": ["mike", "sukmanowsky"],
            },
            {
                "id": "3",
                "comp": "word-spout",
                "stream": "default",
                "task": 0,
                "tuple": ["lida", "elias"],
            },
            {
                "id": "4",
                "comp": "word-spout",
                "stream": "default",
                "task": 0,
                "tuple": ["lida", "smith"],
            },
        )
        for message in messages:
            ShellBatchingBoltTester.shell_proc.write_message(message)

        time.sleep(2.5)

        results = []
        # should have series of emits, then series of acks repeated twice
        for i in xrange(2*2*2):
            results.append(ShellBatchingBoltTester.shell_proc.read_message())

        expected_commands = ["emit"] * 2
        expected_commands += ["ack"] * 2
        expected_commands += ["emit"] * 2
        expected_commands += ["ack"] * 2
        for i, command in enumerate(expected_commands):
            res = results[i]
            self.assertEqual(res.get("command"), command)

    def test_5_worker_exception(self):
        msg = {
            "id": "0",
            "comp": "word-spout",
            "stream": "default",
            "task": 0,
            "tuple": ["fail"],
        }
        ShellBatchingBoltTester.shell_proc.write_message(msg)
        res = ShellBatchingBoltTester.shell_proc.read_message()
        self.assertEqual(res.get("command"), "log")
        self.assertIn("Exception", res.get("msg"))

        res = ShellBatchingBoltTester.shell_proc.read_message()
        self.assertEqual(res.get("command"), "sync")

        self.assertRaises(Exception,
                          ShellBatchingBoltTester.shell_proc.read_message)

    @classmethod
    def tearDownClass(cls):
        os.remove(here(str(cls.proc.pid)))
        cls.proc.kill()


if __name__ == '__main__':
    unittest.main()



import json


class ShellProcess(object):
    """Immitation of https://github.com/apache/incubator-storm/blob/0.8.2/src/jvm/backtype/storm/utils/ShellProcess.java
    """
    DEFAULT_STREAM_ID = 1

    def __init__(self, in_buf, out_buf):
        self.in_buf = in_buf
        self.out_buf = out_buf

    def write_message(self, msg):
        self.write_string(json.dumps(msg))

    def write_string(self, string):
        self.out_buf.write("{}\nend\n".format(string))
        self.out_buf.flush()

    def read_message(self):
        return json.loads(self.read_string())

    def read_string(self):
        res = ""
        while True:
            in_line = self.in_buf.readline()[0:-1]
            if not in_line:
                raise Exception("Pipe to subprocess seems to be broken!")

            if in_line == "end":
                break
            if len(res) != 0:
                res = "{}\n".format(res)
            res = "{}{}".format(res, in_line)

        return res

    def query_subprocess(self, query):
        actions = []
        self.write_message(query)

        while True:
            action = self.read_message()
            actions.append(action)
            if action.get("command") == "sync":
                return actions
            elif action.get("command") == "emit":
                stream = action.get("stream") or self.DEFAULT_STREAM_ID
                task = action.get("task")
                need_task_ids = action.get("need_task_ids")
                if task is None:
                    if need_task_ids is None and need_task_ids == True:
                        self.write_message([1, 2])  # made up task IDs

